## This file is part of kamene
## See http://www.secdev.org/projects/scapy for more informations
## Copyright (C) Philippe Biondi <phil@secdev.org>
## This program is published under a GPLv2 license

"""
Scapy: create, send, sniff, dissect and manipulate network packets.

"""

if __name__ == "__main__":
    from kamene.main import interact
    interact()
