## This file is part of Scapy
## See http://www.secdev.org/projects/scapy for more informations
## Copyright (C) Philippe Biondi <phil@secdev.org>
## This program is published under a GPLv2 license

"""
Support for BSD-like operating systems such as FreeBSD, OpenBSD and Mac OS X.
"""

LOOPBACK_NAME="lo0"

from .unix import *
